# Copyright (c) 2016 - for information on the respective copyright owner
# see the NOTICE file and/or the repository https://github.com/boschresearch/assets2036py.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging
import ssl
import time
import traceback
import typing
from abc import ABC, abstractmethod
from enum import Enum
from json import JSONDecodeError
from numbers import Number
from os import path

from typing import Any, Callable, Union

from urllib.request import urlopen

import jsonschema

from assets2036py.exceptions import InvalidParameterException, NotWritableError
from .communication import CommunicationClient
from .utilities import sanitize, get_resource_path

# pylint: disable=protected-access, line-too-long


logger = logging.getLogger(__name__)
schema_file = path.abspath(path.dirname(__file__)) + "/resources/submodel_schema.json"

with open(schema_file) as f:
    submodel_schema = json.load(f)


# pylint: disable=no-member


class PropertyType(Enum):
    BOOLEAN = 0
    INTEGER = 1
    NUMBER = 2
    STRING = 3
    OBJECT = 5
    NULL = 6
    ARRAY = 7

    def native_type(self):
        type_map = {
            self.BOOLEAN: bool,
            self.INTEGER: int,
            self.NUMBER: Number,
            self.OBJECT: dict,
            self.STRING: str,
            self.ARRAY: list
        }
        if self in type_map:
            return type_map[self]
        else:
            raise Exception(f"No native Type for {self.name}")


class Mode(Enum):
    OWNER = 0
    CONSUMER = 1


class Property(ABC):
    def __init__(self, name, parent, property_definition):
        self._val = None
        self.schema = property_definition
        self.property_type = PropertyType[property_definition["type"].upper()]
        self.parent = parent
        self.communication_client = parent.communication_client
        self.name = name
        self.callbacks = []

    # noinspection PyProtectedMember

    def _get_topic(self):
        return self.parent._get_topic() + f"/{self.name}"

    def assert_valid_type(self, value):
        jsonschema.validate(value, self.schema)

    @property
    @abstractmethod
    def value(self):
        return self._val

    @value.setter
    @abstractmethod
    def value(self, value):
        pass

    def on_change(self, callback):
        self.callbacks.append(callback)


class WritableProperty(Property):

    @property
    def value(self):
        return self._val

    @value.setter
    def value(self, value):
        self.assert_valid_type(value)
        self._val = value
        self.communication_client.publish(self._get_topic(), json.dumps(value))

    def delete(self):
        logger.debug("deleting %s", self.name)
        self.communication_client.publish(self._get_topic(), "")


class ReadOnlyProperty(Property):

    def __init__(self, name, parent, property_definition):
        super().__init__(name, parent, property_definition)
        self.communication_client.subscribe(self._get_topic(), self._update)

    @property
    def value(self):
        return self._val

    @value.setter
    def value(self, value):
        raise NotWritableError("Writing to a ReadonlyProperty not allowed.")

    def _update(self, payload):
        logger.debug("Updating %s to %s", self.name, payload)
        try:
            updated_value = json.loads(payload)
            self.assert_valid_type(updated_value)
            self._val = updated_value
            for cb in self.callbacks:
                cb(self._val)
        except JSONDecodeError as jsonexc:
            logger.error("Payload '%s' was invalid JSON: %s", payload, jsonexc)
        except AssertionError as assertexc:
            logger.error("Parsed payload '%s' has wrong type: %s", payload, assertexc)


class Event(ABC):
    def __init__(self, name, parent, event_definition):
        self.name = name
        self.parent = parent
        self.communication_client = parent.communication_client
        self._parameters = {}
        if "parameters" in event_definition:
            self._parameters = {name: PropertyType[schema["type"].upper()] for name, schema in
                                event_definition["parameters"].items()}

    def _validate_parameters(self, params):
        if params.keys() == self._parameters.keys():
            return all(isinstance(val, self._parameters[name].native_type()) for name, val in params.items())
        return False

    def _get_topic(self):
        return self.parent._get_topic() + f"/{self.name}"


class SubscribableEvent(Event):

    def on_event(self, callback):
        def callback_func(parameters, timestamp):
            if not self._validate_parameters(parameters):
                raise InvalidParameterException(
                    f" expected {','.join(self._parameters.keys())} but received {parameters}")
            callback(timestamp, **parameters)

        self.communication_client.subscribe_event(self._get_topic(), callback_func)


class TriggerableEvent(Event):

    def trigger(self, **params):
        if not self._validate_parameters(params):
            raise InvalidParameterException(f" expected {self._parameters.keys()} but got {params}")
        self.communication_client.trigger_event(self._get_topic(), params)


class Operation(ABC):
    def __init__(self, name, parent, operation_definition):
        self.name = name
        self.parent = parent
        self.communication_client = parent.communication_client
        self._response_schema = None
        self._parameters = {}
        if "parameters" in operation_definition:
            self._parameters = {name: PropertyType[schema["type"].upper()] for name, schema in
                                operation_definition["parameters"].items()}
        if "response" in operation_definition:
            self._response_type = PropertyType[operation_definition["response"]["type"].upper()]
            self._response_schema = operation_definition["response"]

    @abstractmethod
    def invoke(self, timeout=None, **params):
        raise NotImplementedError("You can only call remote operations..")

    def _validate_parameters(self, params):
        if params.keys() == self._parameters.keys():
            return all(isinstance(val, self._parameters[name].native_type()) for name, val in params.items())
        return False

    # noinspection PyProtectedMember
    def _get_topic(self):
        return self.parent._get_topic() + f"/{self.name}"


class CallableOperation(Operation):

    def invoke(self, timeout=None, **params):
        if not self._validate_parameters(params):
            raise InvalidParameterException(f" expected {self._parameters} but got {params}")
        logger.debug("Calling %s", self.name)
        res = self.communication_client.invoke_operation(self._get_topic(), params, timeout)
        logger.debug("%s got %s", self.name, res)
        if hasattr(self, "_response_type"):
            assert isinstance(res, self._response_type.native_type())
            return res


class BindableOperation(Operation):
    def invoke(self, timeout=None, **params):
        super().invoke(timeout=timeout, **params)

    def bind(self, callback):
        def callback_func(parameters):
            if not self._validate_parameters(parameters):
                raise InvalidParameterException(
                    f" expected list with {self._parameters.keys()} but received {parameters}")
            try:
                res = callback(**parameters)
                logger.debug("callback executed, response is %s", res)
            # pylint: disable=broad-except
            except Exception:
                logger.error("operation '%s' caused exception:\n%s", callback, traceback.format_exc())
            if hasattr(self, "_response_type"):
                jsonschema.validate(res, self._response_schema)
                return res

        self.communication_client.bind_operation(self._get_topic(), callback_func)


class SubModel:
    meta_property = {
        "type": "object",
        "properties": {
            "source": {"type": "string"},
            "submodel_schema": {"type": "object"},
            "submodel_url": {"type": "string"}}
    }

    def __init__(self, asset, submodel_definition, lazy_loading=False):
        # pylint: disable=no-member
        self.endpoint_asset = None  # will be set in `register_source` method.
        self.parent = asset
        self.communication_client = asset.communication_client
        self.name = sanitize(submodel_definition["name"])
        self.submodel_definition = submodel_definition
        self._disconnect_callback = None

        if asset.access_mode == Mode.OWNER:
            meta_prop = WritableProperty("_meta", self, self.meta_property)
            setattr(self, "_meta", meta_prop)

            if "operations" in submodel_definition:
                self._create_bindable_operations(submodel_definition)
        else:
            if "operations" in submodel_definition:
                self._create_callable_operations(submodel_definition)

        if not lazy_loading and "properties" in submodel_definition:
            for name, schema in submodel_definition["properties"].items():
                self._create_property_attribute(name, schema, asset.access_mode)

    def __getattr__(self, name: str):
        """
        Overwritten __getattr__ method. Instantiates and returns event objects to avoid heavy load on startup.

        Raises:
            AttributeError: If name is not specified in submodel definition.
        """
        # Note for further developers: If you need to access attributes of your own instance, where you are not sure
        # if they are already set, then use: object.__getattribute__(self, name) to prevent endless loop creation.

        # Get the access mode from the parent node
        access_mode = self.parent.access_mode
        # Check if item is in "events"
        if "events" in self.submodel_definition and name in self.submodel_definition["events"]:
            # requested item found as event. Create event and return instance
            schema = self.submodel_definition["events"][name]
            new_event = self._create_event_attribute(name, schema, access_mode)
            return new_event
        if "properties" in self.submodel_definition and name in self.submodel_definition["properties"]:
            # requested item found as property
            schema = self.submodel_definition["properties"][name]
            new_prop = self._create_property_attribute(name, schema, access_mode)
            return new_prop
        # We couldn't find the attribute
        raise AttributeError(f"No attribute named {name} found in submodel_definition.")

    def is_online(self, seconds_to_wait: int) -> bool:
        """
        Checks if the endpoint referenced by the submodel is online.
        Checks for at most `seconds_to_wait` until the online status is assumed offline.
        Args:
            seconds_to_wait: Amount of seconds before declaring the device as offline if no answer appeared.

        Returns:
            Online status of the device. False if no answer within `seconds_to_wait`.
        """
        num_tries = 0
        if self.name == "_endpoint":
            online_prop = self.online
        else:
            online_prop = self.endpoint_asset._endpoint.online
        while num_tries < seconds_to_wait * 2:
            online_state = online_prop.value
            if online_state is None:
                num_tries += 1
                time.sleep(0.5)
            else:
                return online_state
        return False

    def _raise_offline_exception(self, online):
        logger.debug("CALLBACK GOT %s", online)
        if not online and self._disconnect_callback is not None:
            logger.debug("SEND DISCONNECT FOR %s", self.name)
            self._disconnect_callback(self.name)

    def on_disconnect(self, callback):
        self._disconnect_callback = callback

    def register_source(self, source_address):
        if "/" in source_address:
            namespace, source_asset = source_address.split("/")
        else:
            namespace = self.parent.namespace
            source_asset = source_address
        if source_asset == self.parent.name and namespace == self.parent.namespace:
            # this asset is endpoint for itself.
            self.endpoint_asset = self.parent
        else:
            with open(get_resource_path("_endpoint.json")) as file:
                endpoint_sm_definition = json.load(file)
            self.endpoint_asset = Asset(source_asset, namespace, endpoint_sm_definition, mode=Mode.CONSUMER,
                                        communication_client=self.communication_client, endpoint_name="")
        self.endpoint_asset._endpoint.online.on_change(self._raise_offline_exception)

    def delete(self):
        deletables = [getattr(self, prop) for prop in dir(self) if type(getattr(self, prop)) == WritableProperty]
        deletables.extend(getattr(self, prop) for prop in dir(self) if type(getattr(self, prop)) == BindableOperation)
        for d in deletables:
            d.delete()

    # noinspection PyProtectedMember
    def _get_topic(self):
        return self.parent._get_topic() + f"/{self.name}"

    def _create_callable_operations(self, submodel_definition):
        for name, schema in submodel_definition["operations"].items():
            new_op = CallableOperation(name, self, schema)
            setattr(self, sanitize(name), new_op.invoke)

    def _create_bindable_operations(self, submodel_definition):
        for name, schema in submodel_definition["operations"].items():
            new_op = BindableOperation(name, self, schema)
            setattr(self, "bind_" + sanitize(name), new_op.bind)


    def _create_property_attribute(self, name: str, schema: dict, mode: Mode)\
            -> Union[ReadOnlyProperty, WritableProperty]:

        """
        Create a property instance and attach it to the submodel as its attribute.
        Args:
            name: property name to be created (and under which the attribute will be attached)
            schema: Schema description of the property, containing e.g. parameters etc.
            mode: Decision to create the property in Consumer (`ReadOnlyProperty`) or Owner (`WritableProperty`) mode

        Returns:
            New property class instance.
        """
        if mode == mode.CONSUMER:
            new_prop = ReadOnlyProperty(name, self, schema)
        elif mode == mode.OWNER:
            new_prop = WritableProperty(name, self, schema)
        else:  # Internal error with unknown mode
            typing.assert_never(mode)
        # Set the attribute in the class instance and return value.
        setattr(self, sanitize(name), new_prop)
        return new_prop


    def _create_event_attribute(self, name: str, schema: dict, mode: Mode) -> Union[SubscribableEvent, Callable]:

        """
        Create an event instance and attach it to the submodel as its attribute.
        Args:
            name: event name to be created (and under which the attribute will be attached)
            schema: Schema description of the event, containing e.g. parameters etc.
            mode: Decision to create an event in Consumer (SubscribableEvent) or Owner (TriggerableEvent) mode

        Returns:
            New event class instance. Can also be reached via self.name afterward.
            For TriggerableEvent, return the `trigger` function directly
        """
        if mode == mode.CONSUMER:
            new_event = SubscribableEvent(name, self, schema)
            setattr(self, sanitize(name), new_event)
            return new_event
        elif mode == mode.OWNER:
            new_event = TriggerableEvent(name, self, schema)
            setattr(self, sanitize(name), new_event.trigger)
            return new_event.trigger
        else:
            # Internal error with unknown mode
            typing.assert_never(mode)


class Asset:
    """Core element of assets2036py. Assets are automatically generated during runtime from given submodel descriptions.

    For each submodel description given an asset receives an attribute of the same name.
    Each submodel then implements the operations, properties and events specified by the submodel.

    You shouldn't instantiate assets by yourself, make use of :class: `assets2036py.assetmanager`

    """

    def __init__(self, name: str, namespace: str, *sub_models: dict, mode: Mode = Mode.CONSUMER,
                 communication_client: CommunicationClient, endpoint_name: str, lazy_loading: bool = False) -> None:
        """
        Initialize the asset object
        Args:
            name: Name of the asset
            namespace: Namespace in which the asset is present
            *sub_models: Tuple of submodel definitions the asset shall implement
            mode: Decider if the asset shall be in consumer or provider mode
            communication_client: Communication client to connect to the asset management service (e.g. MQTT Broker)
            endpoint_name: Endpoint name of the asset, used for health status check.
            lazy_loading: If activated, properties of the implemented submodel are only loaded after first reference.
        """
        self.name = name
        self.endpoint_name = endpoint_name
        self.namespace = namespace
        self.access_mode = mode
        self.communication_client = communication_client
        self.sub_models = sub_models
        self.sub_model_names = []
        self.lazy_loading = lazy_loading
        for sm in sub_models:
            self.implement_sub_model(sm)

    def disconnect(self):
        logger.debug("%s disconnects", self.name)
        self.communication_client.disconnect()

    def delete(self):
        for submodel in self.sub_model_names:
            getattr(self, submodel).delete()

    def implement_sub_model(self, submodel):
        if isinstance(submodel, str):
            try:
                # try parsing as json string
                submodel_def = json.loads(submodel)
                submodel_url = "file://localhost"
            except JSONDecodeError:
                # try to load as a web resource
                try:
                    with urlopen(submodel, context=ssl._create_unverified_context()) as response:
                        submodel_def = json.load(response)
                    submodel_url = submodel
                except ValueError as valexc:
                    logger.error("Could not parse submodel definition of %s:\n%s", submodel, valexc)
                    return
        else:
            submodel_def = submodel
            submodel_url = "file://localhost"
            try:
                jsonschema.validate(submodel_def, submodel_schema)
            except jsonschema.exceptions.ValidationError as valexc:
                logger.error("%s is malformed:\n%s", submodel_def, valexc)

        new_sm = SubModel(self, submodel_def, lazy_loading=self.lazy_loading)
        if self.access_mode == Mode.OWNER:
            new_sm._meta.value = {"source": f"{self.namespace}/{self.endpoint_name}",
                                  "submodel_definition": submodel_def,
                                  "submodel_url": submodel_url if submodel_url else "file://localhost"}
        submodel_name = submodel_def['name'].replace("-", "_").replace(" ", "")
        self.sub_model_names.append(submodel_name)
        setattr(self, submodel_name, new_sm)

    def _get_topic(self):
        return f"{self.namespace}/{self.name}"


class ProxyAsset(Asset):

    def __init__(self, name: str, namespace: str, *meta_infos: dict[str, Any], mode: Mode = Mode.CONSUMER,
                 communication_client: CommunicationClient, endpoint_name: str, lazy_loading: bool = False) -> None:
        """
        Initialize the ProxyAsset object
        Args:
            lazy_loading:
            name: Name of the asset
            namespace: Namespace in which the asset is present
            *meta_infos: Tuple of metadata information for the proxy
            mode: Decider if the asset shall be in consumer or provider mode
            communication_client: Communication client to connect to the asset management service (e.g. MQTT Broker)
            endpoint_name: Endpoint name of the asset, used for health status check.
            lazy_loading: If activated, properties of the implemented submodel are only loaded after first reference.
        """
        self.timeout_online = 3
        """ Time to wait before declaring device as offline"""
        submodel_defs = []
        sources = {}
        for meta_info in meta_infos:
            sm_def = meta_info["submodel_definition"]
            submodel_defs.append(sm_def)
            sources[sm_def["name"]] = meta_info["source"]

        super().__init__(name, namespace, *submodel_defs, mode=mode, communication_client=communication_client,
                         endpoint_name=endpoint_name, lazy_loading=lazy_loading)
        for name, source in sources.items():
            if name != "_endpoint":
                getattr(self, name).register_source(source)
                # this is some hacky shit, we need some nicer refactoring here

    @property
    def is_online(self) -> bool:
        """
        True if all submodels of the proxy asset are online.
        Adapt wait time by adapting `self.timeout_online`.
        """
        return self.wait_for_online(self.timeout_online)

    def wait_for_online(self, seconds_to_wait: int) -> bool:
        """
        Blocks for `seconds_to_wait` or until an online_status is received. If timeout appears, return False.
        Args:
            seconds_to_wait: Wait time for answer until declaring offline state

        Returns:
            True/False depending on device status.
        """
        for submodel in self.sub_model_names:
            if not getattr(self, submodel).is_online(seconds_to_wait):
                return False
        return True

    def on_disconnect(self, callback):
        for submodel in self.sub_model_names:
            getattr(self, submodel).on_disconnect(callback)



# Copyright (c) 2016 - for information on the respective copyright owner
# see the NOTICE file and/or the repository https://github.com/boschresearch/assets2036py.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import time
from os import path
from urllib.request import pathname2url

import paho.mqtt.client as mqtt

logger = logging.getLogger(__name__)

res_path = path.abspath(path.dirname(__file__)) + "/resources/"
res_url = "file:" + pathname2url(res_path)
HOST = os.getenv("MQTT_BROKER_URL", "localhost")


def get_msgs_for_n_secs(topic, seconds, host=HOST):
    msgs = []

    def _message_callback(_client, _userdata, message):
        logger.debug("Callback received %s on %s", message.payload, message.topic)
        msgs.append(message)

    def _on_connect(client, _userdata, _flags, _reason_code, _properties):
        client.subscribe(topic)

    client = mqtt.Client(mqtt.CallbackAPIVersion.VERSION2)
    client.on_message = _message_callback
    client.on_connect = _on_connect
    client.connect(host)
    client.loop_start()
    time.sleep(seconds)
    client.loop_stop()
    client.disconnect()
    return msgs


def wipe_retained_msgs(host, namespace=None, seconds=3):
    def _message_callback(client, _userdata, message):
        if message.payload != b"":
            logger.debug("Wiping %s from %s", message.payload, message.topic)
            client.publish(message.topic, "", retain=True)

    def _on_connect(client, _userdata, _flags, _reason_code, _properties):
        topic = namespace + "/#" if namespace else "#"
        logger.debug("subscribing to %s", topic)
        client.subscribe(topic)

    client = mqtt.Client(mqtt.CallbackAPIVersion.VERSION2)
    client.on_message = _message_callback
    client.on_connect = _on_connect
    client.connect(host)
    client.loop_start()
    time.sleep(seconds)
    client.loop_stop()
    client.disconnect()



# Copyright (c) 2016 - for information on the respective copyright owner
# see the NOTICE file and/or the repository https://github.com/boschresearch/assets2036py.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging
import os
import threading
import time
from unittest import TestCase

from assets2036py import AssetManager

from .test_utils import get_msgs_for_n_secs, wipe_retained_msgs

logger = logging.getLogger(__name__)

HOST = os.getenv("MQTT_BROKER_URL", "localhost")
PORT = int(os.getenv("MQTT_BROKER_PORT", "1883"))

NAMESPACE = "test_assetmgr_on_off_arena2036"


class TestAssetManager(TestCase):
    @classmethod
    def tearDownClass(cls):
        wipe_retained_msgs(HOST, NAMESPACE)

    def setUp(self):
        wipe_retained_msgs(HOST, NAMESPACE)
        time.sleep(2)

    def test_on_off_detection(self):
        mgr1 = AssetManager(HOST, PORT, NAMESPACE, "mgr1")
        mgr2 = AssetManager(HOST, PORT, NAMESPACE, "mgr2")
        mgr3 = AssetManager(HOST, PORT, NAMESPACE, "mgr3")

        _asset1 = mgr1.create_asset(
            "test_asset_123",
            "https://raw.githubusercontent.com/boschresearch/assets2036-submodels/master/location.json",
        )
        _asset2 = mgr2.create_asset(
            "test_asset_123",
            "https://raw.githubusercontent.com/boschresearch/assets2036-submodels/master/location.json",
        )
        _asset3 = mgr3.create_asset(
            "test_asset_123",
            "https://raw.githubusercontent.com/boschresearch/assets2036-submodels/master/location.json",
        )


        time.sleep(2)
        msgs = get_msgs_for_n_secs(f"{NAMESPACE}/+/_endpoint/online/#", 2)
        logger.debug([msg.payload for msg in msgs])
        endpoints = {msg.topic: json.loads(msg.payload) for msg in msgs}

        self.assertEqual(len(msgs), 3)
        self.assertTrue(all(endpoints.values()))

        mgr2.disconnect()
        time.sleep(1)

        msgs = get_msgs_for_n_secs(f"{NAMESPACE}/+/_endpoint/online/#", 2)
        endpoints = {msg.topic: json.loads(msg.payload) for msg in msgs}

        self.assertEqual(len(msgs), 3)
        self.assertFalse(all(endpoints.values()))

        mgr1.client.disconnect()
        mgr3.client.disconnect()
        time.sleep(1)
        msgs = get_msgs_for_n_secs(f"{NAMESPACE}/+/_endpoint/online/#", 2)
        endpoints = {msg.topic: json.loads(msg.payload) for msg in msgs}

        self.assertEqual(len(msgs), 3)
        self.assertTrue(all(not v for v in endpoints.values()))

    def test_healthy_flag(self):
        mgr1 = AssetManager(HOST, PORT, NAMESPACE, "mgr1")

        _asset = mgr1.create_asset(
            "test_asset_123",
            "https://raw.githubusercontent.com/boschresearch/assets2036-submodels/master/location.json",
        )

        msgs = get_msgs_for_n_secs(f"{NAMESPACE}/mgr1/_endpoint/healthy/#", 2)
        self.assertTrue(len(msgs), 1)
        self.assertEqual(msgs[0].payload, b"false")
        logger.debug(msgs)

        mgr1.healthy = True

        msgs2 = get_msgs_for_n_secs(f"{NAMESPACE}/mgr1/_endpoint/healthy/#", 2)
        self.assertTrue(len(msgs2), 1)
        self.assertEqual(msgs2[0].payload, b"true")

    def test_query_assets(self):
        mgr = AssetManager(HOST, PORT, NAMESPACE, "mgr")
        endpoints = mgr.query_assets("arena2036", "_endpoint")
        self.assertTrue(len(endpoints) > 0)

    def test_query_all(self):
        mgr = AssetManager(HOST, PORT, NAMESPACE, "mgr")
        assets = mgr.query_assets("+", "+")
        self.assertTrue(len(assets) > 0)

    def test_create_asset_proxy(self):
        mgr = AssetManager(HOST, PORT, NAMESPACE, "mgr")
        _asset = mgr.create_asset(
            "test_asset_123",
            "https://arena2036-infrastructure.saz.bosch-si.com/arena2036_public/assets2036_submodels/raw/master/powerstate.json",
        )
        asset_proxy = mgr.create_asset_proxy(NAMESPACE, "test_asset_123")
        self.assertIn("powerstate", dir(asset_proxy))

    def test_self_ping(self):
        # pylint: disable=protected-access
        mgr = AssetManager(HOST, PORT, NAMESPACE, "mgr")
        _asset = mgr.create_asset(
            "test_asset_123",
            "https://arena2036-infrastructure.saz.bosch-si.com/arena2036_public/assets2036_submodels/raw/master/powerstate.json",
        )
        self.assertTrue(mgr._self_ping())
        mgr.disconnect()
        self.assertFalse(mgr._self_ping())

    def test_shutdown(self):
        mgr = AssetManager(HOST, PORT, NAMESPACE, "mgr")
        _asset = mgr.create_asset(
            "test_asset_123",
            "https://arena2036-infrastructure.saz.bosch-si.com/arena2036_public/assets2036_submodels/raw/master/powerstate.json",
        )
        mgr.set_healthy_callback(lambda: True)
        monitor_threads = [t for t in threading.enumerate() if t.name[:3] == "mgr"]
        self.assertEqual(len(monitor_threads), 1)
        monitor_thread = monitor_threads[0]
        self.assertTrue(monitor_thread.is_alive())
        mgr.disconnect()
        time.sleep(3)
        self.assertFalse(monitor_thread.is_alive())

    def test_logging_handler(self):
        _logger = logging.getLogger("testing_logger")
        mgr = AssetManager(HOST, PORT, NAMESPACE, "mgr")
        lh = mgr.get_logging_handler()
        _logger.addHandler(lh)
        _logger.setLevel(logging.DEBUG)
        _logger.debug("Hallo Test")



import logging
from .assets import SubModel
from logging import LogRecord


class AssetLoggingHandler(logging.Handler):
    """
    Custom Logging handler class to log via MQTT using the _endpoint of an asset.
    """
    def __init__(self, endpoint: SubModel, level: int = 0) -> None:
        super().__init__(level)
        self._endpoint = endpoint

    def emit(self, record: LogRecord) -> None:
        self._endpoint.log(entry=record.msg)



# Copyright (c) 2016 - for information on the respective copyright owner
# see the NOTICE file and/or the repository https://github.com/boschresearch/assets2036py.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging
import os
import ssl
from threading import Event, Thread
from typing import Type
from urllib.request import urlopen

from assets2036py import Asset, Mode, ProxyAsset
from assets2036py.assetlogging import AssetLoggingHandler
from assets2036py.communication import MQTTClient
from assets2036py.exceptions import (
    AssetNotFoundError,
    AssetNotOnlineError,
    OperationTimeoutException,
)
from assets2036py.utilities import get_resource_path

logger = logging.getLogger(__name__)


# pylint: disable=line-too-long


class AssetManager:
    """Primary entrypoint for working with assets.
    Manages the creation of assets and asset proxies (the latter are used to talk to remote assets).

    """

    def __init__(
        self,
        host: str,
        port: int,
        namespace: str,
        endpoint_name: str,
        communication_client: Type[MQTTClient] = MQTTClient,
    ) -> None:
        """
        Args:
            host (str): URL of MQTT broker or registry
            port (int): Port of MQTT broker or registry
            namespace (str): default namespace to use for asset creation
            endpoint_name (str): name of the endpoint to use
            communication_client: Backend to use for communication. Defaults to MQTTClient.
        """
        self.endpoint_name = endpoint_name
        self.port = port
        self.host = host
        self.namespace = namespace
        self.client = communication_client(client_id=endpoint_name)
        self.client.connect(
            host=host, port=port, namespace=namespace, endpoint_name=endpoint_name
        )
        self._endpoint = None
        self._shutdown = Event()

    def __del__(self):
        """
        Destructor definition. Disconnect underlying communication client.
        """
        self.disconnect()

    def disconnect(self):
        """disconnect underlying communication client"""
        self._shutdown.set()
        self.client.disconnect()

    def get_logging_handler(self):
        """
        Return a logging handler which enables to log onto MQTT via the endpoint of the assetmanager.

        Examples:
            For registering a logger, follow these steps:

            _logger = logging.getLogger("testing_logger")
            mgr = AssetManager(HOST, PORT, NAMESPACE, "mgr")
            lh = mgr.get_logging_handler()
            _logger.addHandler(lh)
            _logger.info("Hallo Test")
        """
        if not self._endpoint:
            self._create_endpoint_asset()
        return AssetLoggingHandler(self._endpoint._endpoint)

    @property
    def healthy(self) -> bool:
        """get/set healthy state of endpoint

        Returns:
            bool: healthy flag state
        """
        # pylint: disable=protected-access
        if self._endpoint:
            return self._endpoint._endpoint.healthy.value
        else:
            return False

    @healthy.setter
    def healthy(self, val: bool):
        """set health state

        Args:
            val (bool):  healthy flag state
        """
        # pylint: disable=protected-access
        if self._endpoint:
            self._endpoint._endpoint.healthy.value = val

    def set_healthy_callback(
        self, callback: callable, interval: int = 30, kill_on_unhealthy: bool = False
    ):
        """Regularly calls the given callback and sets the healthy state accordingly

        Optionally kills the complete application if healthy resolves to False

        Args:
            callback (function): callback that will be called each <interval> seconds. Receives no parameters, must return True or False
            interval (int, optional): callback will be called every <interval> seconds. Defaults to 30.
            kill_on_unhealthy (bool, optional): If set to True, os._exit(-1) is called and app is terminated. Defaults to False.
        """

        def monitor_loop():
            while not self._shutdown.is_set():
                self.healthy = self._self_ping() and callback()
                if kill_on_unhealthy and not self.healthy:
                    # pylint: disable=protected-access
                    os._exit(-1)
                self._shutdown.wait(interval)

        Thread(target=monitor_loop, name=f"{self.endpoint_name}_Monitor").start()

    def _self_ping(self) -> bool:
        """internal helper to check if connection to broker is still up
        Returns:
            bool: True if ping successfully, else False
        """
        if not self._endpoint:
            return False
        try:
            proxy_self = self.create_asset_proxy(self.namespace, self.endpoint_name)
            # pylint: disable=protected-access
            proxy_self._endpoint.ping(timeout=2)
            return True
        except (OperationTimeoutException, AssetNotFoundError) as ote:
            logger.error("self ping failed: %s", ote)
            return False

    def _create_endpoint_asset(self) -> None:
        """create the endpoint asset (for internal use)"""
        with open(get_resource_path("_endpoint.json")) as file:
            endpoint_sm_definition = json.load(file)
        endpoint = self.create_asset(
            self.endpoint_name, endpoint_sm_definition, create_endpoint=False
        )

        # pylint: disable=protected-access
        endpoint._endpoint.online.value = True
        endpoint._endpoint.healthy.value = False
        endpoint._endpoint.bind_shutdown(self.shutdown)
        endpoint._endpoint.bind_restart(self.restart)
        endpoint._endpoint.bind_ping(lambda: None)

        # quick hack
        def set_online(_client, _userdata, _flags, _reason_code, properties):
            endpoint._endpoint.online.value = True

        self.client.on_connect(set_online)

        self._endpoint = endpoint

    def shutdown(self):
        """Disconnect from broker and shut down (gracelessly)"""
        logger.debug("Shutting down.")
        self.disconnect()
        # pylint: disable=protected-access
        os._exit(0)

    def restart(self):
        """placeholder for endpoint operation "restart" """
        logger.debug("Restarting. TBI")
        self.shutdown()

    def create_asset(
        self,
        name: str,
        *sub_models: str,
        mode=Mode.OWNER,
        namespace: str = None,
        create_endpoint: bool = True,
        lazy_loading: bool = False,
    ) -> Asset:
        """Create a new Asset

        Args:
            name (str): Name of the asset to create
            mode (int, optional): set mode to consumer or owner. Defaults to Mode.OWNER.
            namespace (str, optional): Namespace in which asset is created. If not set, namespace of asset manager is used. Defaults to None.
            create_endpoint (bool): Decision if a new endpoint asset shall be created.
            lazy_loading (bool): If activated, properties of the implemented submodel are only loaded after first reference.


        Returns:
            Asset: Newly created asset
        """

        if not namespace:
            namespace = self.namespace
        asset = Asset(
            name,
            namespace,
            *sub_models,
            mode=mode,
            communication_client=self.client,
            endpoint_name=self.endpoint_name,
            lazy_loading=lazy_loading,
        )
        if create_endpoint and not self._endpoint:
            self._create_endpoint_asset()
        return asset

    def join(self, timeout=None):
        """Wait for the background tasks to complete"""
        self.client.join(timeout)

    # pylint: disable=keyword-arg-before-vararg
    def query_assets(self, namespace=None, *submodel_names: str) -> set:
        """return a set of those assets that implement all the given submodels

        Args:
            namespace (str, optional): Namespace to search. Defaults to None.

        Returns:
            list: list of found asset names
        """

        return self.client.query_asset_names(namespace, *submodel_names)

    def create_asset_proxy(
        self,
        namespace: str,
        name: str,
        wait_seconds_until_online: int = -1,
        lazy_loading: bool = False,
    ) -> ProxyAsset:
        """Returns the asset with the given name if found, raises AssetNotFoundError otherwise

        Args:
            namespace (str): namespace to use
            name (str): name of asset to use
            wait_seconds_until_online (int): seconds to wait for the asset to register as online.
                -1 ignores online state and initiates asset (default: -1)
            lazy_loading: If activated, properties of the implemented submodel are only loaded after first reference.

        Raises:
            AssetNotFoundError: raised if Asset cannot be found
            AssetNotOnlineError: raised if Asset doesn't come online in set time

        Returns:
            Asset: returns Asset proxy to be used to communicate with remote asset
        """
        submodels = self.client.query_submodels_for_asset(namespace, name)
        if not submodels:
            raise AssetNotFoundError
        # verified_submodels = self._get_submodel_schemata(submodels)

        proxy_asset = ProxyAsset(
            name,
            namespace,
            *submodels.values(),
            mode=Mode.CONSUMER,
            communication_client=self.client,
            endpoint_name=self.endpoint_name,
            lazy_loading=lazy_loading,
        )
        if wait_seconds_until_online >= 0:
            online = proxy_asset.wait_for_online(wait_seconds_until_online)
            if not online:
                raise AssetNotOnlineError
        return proxy_asset

    def _get_submodel_schemata(self, sub_models):
        # pylint: disable=protected-access,broad-except
        schemata = []
        for meta in sub_models.values():
            try:
                url = meta["submodel_url"]
                with urlopen(url, context=ssl._create_unverified_context()) as response:
                    schemata.append(json.load(response))
            except Exception as loadexc:
                logger.warning('Can\'t load schema from url "%s": %s', url, loadexc)
                try:
                    schemata.append(meta["submodel_definition"])
                    logger.debug("Using schema from submodel_definition")
                except Exception as missingexc:
                    logger.error("No submodel schema retrievable: %s", missingexc)
        return schemata



# Copyright (c) 2016 - for information on the respective copyright owner
# see the NOTICE file and/or the repository https://github.com/boschresearch/assets2036py.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from os import path

from setuptools import find_packages, setup

here = path.abspath(path.dirname(__file__))


with open(path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="assets2036py",
    version="0.1.6",
    url="https://github.com/boschresearch/assets2036spy",
    license="BIOS",
    author="Daniel Ewert (CR/APT5)",
    author_email="Daniel.Ewert@de.bosch.com",
    description="helper library to easily implement assets2036",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(exclude=["contrib", "docs", "tests", "examples"]),
    install_requires=["jsonschema", "paho-mqtt", "python-dateutil"],
    package_data={"assets2036py": ["resources/*.json", "py.typed"]},
)



# Copyright (c) 2016 - for information on the respective copyright owner
# see the NOTICE file and/or the repository https://github.com/boschresearch/assets2036py.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import time

from multiprocessing import Pool
from threading import Thread
from unittest import TestCase

from jsonschema import ValidationError


from assets2036py import Asset, AssetManager, Mode
from assets2036py.communication import MockClient
from assets2036py.exceptions import AssetNotOnlineError, NotWritableError

from .test_utils import get_msgs_for_n_secs, res_url, wipe_retained_msgs


logger = logging.getLogger(__name__)

HOST = os.getenv("MQTT_BROKER_URL", "localhost")
PORT = int(os.getenv("MQTT_BROKER_PORT", "1883"))


# HINT: Add an .env file to the root of your project to set the environment variables


# pylint: disable=protected-access


def create_client_and_ask(idx):

    """Helper function to test parallel operations"""
    mgr = AssetManager(HOST, PORT, "test_arena2036", f"test_endpoint_{idx}")
    client = mgr.create_asset(
        "source", res_url + "simple_operation.json", mode=Mode.CONSUMER
    )

    return client.simple_operation.addnumbers(a=0, b=idx)


class TestAsset(TestCase):
    @classmethod
    def tearDownClass(cls):
        wipe_retained_msgs(HOST, "test_arena2036")

    def test_implement_sub_model_consumer(self):

        a = Asset(
            "simpleasset",
            "test_arena2036",
            res_url + "simplebool.json",
            mode=Mode.CONSUMER,
            communication_client=MockClient(),
            endpoint_name="test_endpoint_1",
        )

        self.assertTrue("simple_bool" in dir(a))
        self.assertTrue("switch" in dir(a.simple_bool))
        self.assertTrue("value" in dir(a.simple_bool.switch))
        topic = a.simple_bool.switch._get_topic()
        self.assertEqual(topic, "test_arena2036/simpleasset/simple_bool/switch")

    def test_implement_sub_model_owner(self):

        a = Asset(
            "simpleasset",
            "test_arena2036",
            res_url + "simplebool.json",
            mode=Mode.OWNER,
            communication_client=MockClient(),
            endpoint_name="test_endpoint_1",
        )

        self.assertTrue("simple_bool" in dir(a))
        self.assertTrue("switch" in dir(a.simple_bool))
        self.assertTrue("value" in dir(a.simple_bool.switch))
        topic = a.simple_bool.switch._get_topic()
        self.assertEqual(topic, "test_arena2036/simpleasset/simple_bool/switch")

    def test_put_value(self):

        a = Asset(
            "simpleasset",
            "test_arena2036",
            res_url + "simplebool.json",
            mode=Mode.OWNER,
            communication_client=MockClient(),
            endpoint_name="test_endpoint_1",
        )

        a.simple_bool.switch.value = True
        self.assertEqual(a.simple_bool.switch.value, True)
        a.simple_bool.switch.value = False
        self.assertEqual(a.simple_bool.switch.value, False)

    def test_load_from_unverified_ssl(self):

        mgr = AssetManager(HOST, PORT, "test_arena2036", "test_endpoint_1")
        source = mgr.create_asset(
            "source",
            "https://arena2036-infrastructure.saz.bosch-si.com/arena2036_public/assets2036_submodels/raw/master/powerstate.json",
        )
        source.powerstate.power_state.value = True

    def test_implement_complex_sub_model_consumer(self):
        a = Asset(
            "activeshuttle",
            "test_arena2036",
            res_url + "activeshuttle.json",
            mode=Mode.CONSUMER,
            communication_client=MockClient(),
            endpoint_name="test_endpoint_1",
        )


        self.assertTrue("active_shuttle" in dir(a))
        self.assertTrue("pose" in dir(a.active_shuttle))
        self.assertTrue("state" in dir(a.active_shuttle))
        self.assertTrue("waypoints" in dir(a.active_shuttle))

        self.assertIsInstance(a.active_shuttle.pose.value, type(None))

        def evil_func():

            a.active_shuttle.pose.value = {"x": 5.7, "y": 3.5, "z": 0.1}


        self.assertRaises(NotWritableError, evil_func)
        self.assertIsInstance(a.active_shuttle.pose.value, type(None))

    def test_all_property_types(self):

        a = Asset(
            "PropertyTest",
            "test_arena2036",
            res_url + "all_property_model.json",
            mode=Mode.OWNER,
            communication_client=MockClient(),
            endpoint_name="test_endpoint_1",
        )

        a.all_properties.a_switch.value = True
        self.assertEqual(a.all_properties.a_switch.value, True)
        a.all_properties.a_float.value = 4.2
        self.assertAlmostEqual(a.all_properties.a_float.value, 4.2, places=5)
        a.all_properties.a_text.value = "Hullo"
        self.assertEqual(a.all_properties.a_text.value, "Hullo")
        a.all_properties.an_object.value = {"x": 5, "y": 6}
        self.assertDictEqual(a.all_properties.an_object.value, {"x": 5, "y": 6})
        a.all_properties.a_list.value = [1, 2, 3, "a"]
        self.assertListEqual(a.all_properties.a_list.value, [1, 2, 3, "a"])

        def evil_func():
            a.all_properties.an_object.value = {"x": 5, "y": "booh"}

        self.assertRaises(ValidationError, evil_func)
        a.all_properties.an_int.value = 42
        self.assertEqual(a.all_properties.an_int.value, 42)

    def test_type_validation(self):

        a = Asset(
            "PropertyTest",
            "test_arena2036",
            res_url + "all_property_model.json",
            mode=Mode.OWNER,
            communication_client=MockClient(),
            endpoint_name="test_endpoint_1",
        )

        with self.assertRaises(ValidationError):
            a.all_properties.a_switch.value = 42
        self.assertEqual(a.all_properties.a_switch.value, None)

        with self.assertRaises(ValidationError):
            a.all_properties.an_int.value = 42.5

        a.all_properties.a_float.value = 42

        with self.assertRaises(ValidationError):
            a.all_properties.a_text.value = 42

        with self.assertRaises(ValidationError):
            a.all_properties.an_object.value = True

        with self.assertRaises(ValidationError):
            a.all_properties.an_object.value = {"x": 5, "z": 6}

        with self.assertRaises(ValidationError):
            a.all_properties.an_int_object.value = {"x": 5.1, "y": 6.0}

        with self.assertRaises(ValidationError):
            a.all_properties.a_list.value = {"x": 5, "z": 6}

    def test_implement_complex_sub_model_owner(self):

        a = Asset(
            "activeshuttle",
            "test_arena2036",
            res_url + "activeshuttle.json",
            mode=Mode.OWNER,
            communication_client=MockClient(),
            endpoint_name="test_endpoint_1",
        )


        self.assertTrue("active_shuttle" in dir(a))
        self.assertTrue("pose" in dir(a.active_shuttle))
        self.assertTrue("state" in dir(a.active_shuttle))
        self.assertTrue("waypoints" in dir(a.active_shuttle))

    def test_communication(self):

        mgr = AssetManager(HOST, PORT, "test_arena2036", "test_endpoint_1")
        source = mgr.create_asset("source", res_url + "all_property_model.json")
        sink = mgr.create_asset(
            "source", res_url + "all_property_model.json", mode=Mode.CONSUMER
        )


        time.sleep(2)
        source.all_properties.a_switch.value = True
        source.all_properties.an_int.value = 42
        source.all_properties.a_float.value = 4.2
        source.all_properties.a_text.value = "Hullo"
        source.all_properties.an_object.value = {"x": 5, "y": 6}
        source.all_properties.a_list.value = [1, 2, 3, 4]

        logger.debug("Sleeping 2 secs")
        time.sleep(2)

        self.assertEqual(True, sink.all_properties.a_switch.value)
        self.assertEqual(42, sink.all_properties.an_int.value)
        self.assertAlmostEqual(4.2, sink.all_properties.a_float.value, places=5)
        self.assertEqual("Hullo", sink.all_properties.a_text.value)
        self.assertDictEqual({"x": 5, "y": 6}, sink.all_properties.an_object.value)
        self.assertListEqual([1, 2, 3, 4], sink.all_properties.a_list.value)
        mgr.disconnect()

    def test_delete(self):

        mgr = AssetManager(HOST, PORT, "test_arena2036", "test_endpoint_1")

        source = mgr.create_asset("deletetest", res_url + "all_property_model.json")

        time.sleep(2)
        source.all_properties.a_switch.value = True
        source.all_properties.an_int.value = 42
        source.all_properties.a_float.value = 4.2
        source.all_properties.a_text.value = "Hullo"
        source.all_properties.an_object.value = {"x": 5, "y": 6}
        source.all_properties.a_list.value = [1, 2, 3, 4]

        time.sleep(2)
        msgs = get_msgs_for_n_secs("test_arena2036/deletetest/#", 3)
        self.assertTrue(len(msgs) > 0)
        source.delete()
        logger.debug("Sleeping 3 secs")
        time.sleep(2)
        msgs = get_msgs_for_n_secs("test_arena2036/deletetest/#", 3)
        logger.debug("Received messages:")
        self.assertEqual(len(msgs), 0)
        mgr.disconnect()

    def test_retained_communication(self):

        mgr = AssetManager(HOST, PORT, "test_arena2036", "test_endpoint_1")

        source = mgr.create_asset("source", res_url + "all_property_model.json")

        time.sleep(2)
        source.all_properties.a_switch.value = True
        source.all_properties.an_int.value = 42
        source.all_properties.a_float.value = 4.2
        source.all_properties.a_text.value = "Hullo"
        source.all_properties.an_object.value = {"x": 5, "y": 6}
        source.all_properties.a_list.value = [1, 2, 3, 4]


        sink = mgr.create_asset(
            "source", res_url + "all_property_model.json", mode=Mode.CONSUMER
        )


        time.sleep(2)

        self.assertEqual(True, sink.all_properties.a_switch.value)
        self.assertEqual(42, sink.all_properties.an_int.value)
        self.assertAlmostEqual(4.2, sink.all_properties.a_float.value, places=5)
        self.assertEqual("Hullo", sink.all_properties.a_text.value)
        self.assertDictEqual({"x": 5, "y": 6}, sink.all_properties.an_object.value)
        self.assertListEqual([1, 2, 3, 4], sink.all_properties.a_list.value)
        mgr.disconnect()

    def test_communication_updates(self):

        mgr = AssetManager(HOST, PORT, "test_arena2036", "test_endpoint_1")
        source = mgr.create_asset(
            "source", res_url + "all_property_model.json", mode=Mode.OWNER
        )
        sink = mgr.create_asset(
            "source", res_url + "all_property_model.json", mode=Mode.CONSUMER
        )


        returns = []

        def cb(x):
            returns.append(x)

        sink.all_properties.an_int.on_change(cb)

        time.sleep(1)

        for idx in range(10):
            source.all_properties.an_int.value = idx
        time.sleep(3)
        logger.debug(returns)
        self.assertAlmostEqual(
            10, len(returns), delta=2
        )  # VerneMQ sends the last received message twice? --> Sometimes 11 msgs...

    def test_submodel_meta_property(self):
        mgr = AssetManager(HOST, PORT, "test_arena2036", "test_endpoint_1")
        asset1 = mgr.create_asset("asset", res_url + "simplebool.json")
        self.assertTrue("simple_bool" in dir(asset1))
        self.assertTrue("_meta" in dir(asset1.simple_bool))
        self.assertEqual(
            "test_arena2036/test_endpoint_1", asset1.simple_bool._meta.value["source"]
        )
        self.assertEqual(
            f"{res_url}simplebool.json", asset1.simple_bool._meta.value["submodel_url"]
        )

        asset2 = mgr.create_asset(
            "asset", res_url + "simplebool.json", mode=Mode.CONSUMER
        )

        self.assertTrue("simple_bool" in dir(asset2))
        self.assertFalse("_meta" in dir(asset2.simple_bool))

    def test_simple_operation(self):

        mgr = AssetManager(HOST, PORT, "test_arena2036", "test_endpoint_1")
        sink = mgr.create_asset(
            "source", res_url + "simple_operation.json", mode=Mode.CONSUMER
        )

        self.assertIn("simple_operation", dir(sink))
        self.assertIn("do_it", dir(sink.simple_operation))
        self.assertTrue(callable(sink.simple_operation.do_it))
        self.assertIn("addnumbers", dir(sink.simple_operation))
        self.assertTrue(callable(sink.simple_operation.addnumbers))


        owner = mgr.create_asset(
            "source", res_url + "simple_operation.json", mode=Mode.OWNER
        )

        self.assertIn("bind_do_it", dir(owner.simple_operation))
        self.assertIn("bind_addnumbers", dir(owner.simple_operation))

        def do_addnumbers(**kwargs):
            return kwargs["a"] + kwargs["b"]

        # noinspection PyUnusedLocal
        def do_it():
            logger.debug("DOOOOING IT")

        def do_concat(**kwargs):
            return kwargs["a"] + kwargs["b"]

        owner.simple_operation.bind_concat(do_concat)
        owner.simple_operation.bind_addnumbers(do_addnumbers)
        owner.simple_operation.bind_do_it(do_it)
        owner.simple_operation.bind_do_it(lambda: logger.debug("DOING IT Lambda_style"))

        time.sleep(1)

        res = sink.simple_operation.addnumbers(a=5, b=6, timeout=3)
        res2 = sink.simple_operation.concat(a="hello", b="world", timeout=3)
        sink.simple_operation.do_it()
        self.assertEqual(11, res)
        self.assertEqual("helloworld", res2)

    def test_multi_parameter_operation(self):

        mgr = AssetManager(HOST, PORT, "test_arena2036", "test_endpoint_1")
        caller = mgr.create_asset(
            "source", res_url + "multi_param_operation.json", mode=Mode.CONSUMER
        )
        owner = mgr.create_asset(
            "source", res_url + "multi_param_operation.json", mode=Mode.OWNER
        )


        def merge_into_list_samename(first, second, third):
            return [first, second, third]

        def merge_into_list_distorted(second, third, first):
            return [first, second, third]

        def merge_into_list_kwargs(**kwargs):
            return [kwargs["first"], kwargs["second"], kwargs["third"]]

        owner.multiparam.bind_multiparam_1(merge_into_list_samename)
        owner.multiparam.bind_multiparam_2(merge_into_list_kwargs)
        owner.multiparam.bind_multiparam_3(merge_into_list_distorted)
        time.sleep(1)

        res1 = caller.multiparam.multiparam_1(
            first=1, second="hello", third={"a": 1}, timeout=3
        )
        res2 = caller.multiparam.multiparam_2(
            first=1, second="hello", third={"a": 1}, timeout=3
        )
        res3 = caller.multiparam.multiparam_3(
            first=1, second="hello", third={"a": 1}, timeout=3
        )
        self.assertEqual([1, "hello", {"a": 1}], res1)
        self.assertEqual([1, "hello", {"a": 1}], res2)
        self.assertEqual([1, "hello", {"a": 1}], res3)

    def test_object_operation(self):

        mgr = AssetManager(HOST, PORT, "test_arena2036", "test_endpoint_1")
        sink = mgr.create_asset(
            "source", res_url + "object_operation.json", mode=Mode.CONSUMER
        )
        owner = mgr.create_asset(
            "source", res_url + "object_operation.json", mode=Mode.OWNER
        )


        def do_split(a):
            return {"firstletter": a[0], "lastletter": a[-1]}

        owner.object_operation.bind_splitstring(do_split)

        time.sleep(1)

        res = sink.object_operation.splitstring(a="IO")
        self.assertEqual({"firstletter": "I", "lastletter": "O"}, res)

    def test_array_operation(self):

        mgr = AssetManager(HOST, PORT, "test_arena2036", "test_endpoint_1")
        sink = mgr.create_asset(
            "source", res_url + "array_operation.json", mode=Mode.CONSUMER
        )
        owner = mgr.create_asset(
            "source", res_url + "array_operation.json", mode=Mode.OWNER
        )


        def do_give_array(length):
            return [1] * length

        owner.array_operation.bind_give_array(do_give_array)

        time.sleep(1)

        res = sink.array_operation.give_array(length=2, timeout=3)
        self.assertEqual([1, 1], res)

    def testMultiOperationRequests(self):

        mgr = AssetManager(HOST, PORT, "test_arena2036", "test_endpoint_1")
        owner = mgr.create_asset(
            "source", res_url + "simple_operation.json", mode=Mode.OWNER
        )

        self.assertIn("bind_do_it", dir(owner.simple_operation))
        self.assertIn("bind_addnumbers", dir(owner.simple_operation))

        def do_addnumbers(a, b):
            return a + b

        # noinspection PyUnusedLocal
        def do_it():
            logger.debug("DOOOOING IT")

        def do_concat(a, b):
            return a + b

        owner.simple_operation.bind_concat(do_concat)
        owner.simple_operation.bind_addnumbers(do_addnumbers)
        owner.simple_operation.bind_do_it(do_it)

        owner.simple_operation.bind_do_it(lambda: logger.debug("DOING IT Lambda_style"))
        consumers = [
            mgr.create_asset(
                "source", res_url + "simple_operation.json", mode=Mode.CONSUMER
            )
            for _ in range(10)
        ]


        for idx, c in enumerate(consumers):
            res = c.simple_operation.addnumbers(a=3 + idx, b=4 + idx)
            self.assertEqual(res, 3 + idx + 4 + idx)

    def testParallelOperationRequests(self):

        mgr = AssetManager(HOST, PORT, "test_arena2036", "test_endpoint_1")
        owner = mgr.create_asset(
            "source", res_url + "simple_operation.json", mode=Mode.OWNER
        )

        self.assertIn("bind_do_it", dir(owner.simple_operation))
        self.assertIn("bind_addnumbers", dir(owner.simple_operation))

        def do_addnumbers(a, b):
            return a + b

        # noinspection PyUnusedLocal
        def do_it():
            logger.debug("DOOOOING IT")

        def do_concat(a, b):
            return a + b

        owner.simple_operation.bind_concat(do_concat)
        owner.simple_operation.bind_addnumbers(do_addnumbers)
        owner.simple_operation.bind_do_it(do_it)
        owner.simple_operation.bind_do_it(lambda: logger.debug("DOING IT Lambda_style"))

        with Pool(10) as p:
            result = p.map(create_client_and_ask, list(range(10)))
            self.assertEqual(result, list(range(10)))

    def testEvents(self):

        mgr = AssetManager(HOST, PORT, "test_arena2036", "test_endpoint_1")
        eventer = mgr.create_asset(
            "eventer", res_url + "simpleevent.json", mode=Mode.OWNER
        )
        listener = mgr.create_asset(
            "eventer", res_url + "simpleevent.json", mode=Mode.CONSUMER
        )


        cb_counter = []

        def bool_cb(ts, **kwargs):
            cb_counter.append(1)

        def string_cb(ts, **kwargs):
            cb_counter.append(1)

        def object_cb(ts, **kwargs):
            cb_counter.append(1)

        listener.simpleevent.bool_happened.on_event(bool_cb)
        listener.simpleevent.string_happened.on_event(string_cb)
        listener.simpleevent.object_happened.on_event(object_cb)

        eventer.simpleevent.bool_happened(bool_payload=True)
        eventer.simpleevent.string_happened(string_payload="it'samee")
        eventer.simpleevent.object_happened(object_payload={"x": 42.1, "y": 3.14159})

        time.sleep(5)
        logger.debug("Waited 5 secs")
        self.assertEqual(len(cb_counter), 3)

    def test_events_named_callbacks(self):

        mgr = AssetManager(HOST, PORT, "test_arena2036", "test_endpoint_1")
        eventer = mgr.create_asset(
            "eventer", res_url + "simpleevent.json", mode=Mode.OWNER
        )

        listener2 = mgr.create_asset(
            "eventer", res_url + "simpleevent.json", mode=Mode.CONSUMER
        )


        cb_counter = []

        # pylint: disable=unused-argument

        def bool_cb_named(_ts, bool_payload):
            cb_counter.append(1)

        def string_cb_named(_ts, string_payload):
            cb_counter.append(1)

        def object_cb_named(_ts, object_payload):
            cb_counter.append(1)

        listener2.simpleevent.bool_happened.on_event(bool_cb_named)
        listener2.simpleevent.string_happened.on_event(string_cb_named)
        listener2.simpleevent.object_happened.on_event(object_cb_named)

        eventer.simpleevent.bool_happened(bool_payload=True)
        eventer.simpleevent.string_happened(string_payload="it'samee")
        eventer.simpleevent.object_happened(object_payload={"x": 42.1, "y": 3.14159})
        time.sleep(3)

        self.assertEqual(len(cb_counter), 3)

    def test_empty_event(self):

        mgr = AssetManager(HOST, PORT, "test_arena2036", "test_endpoint_1")
        eventer = mgr.create_asset(
            "eventer", res_url + "simpleevent.json", mode=Mode.OWNER
        )

        listener2 = mgr.create_asset(
            "eventer", res_url + "simpleevent.json", mode=Mode.CONSUMER
        )

        callback_called = False

        def empty_callback(_ts):
            nonlocal callback_called
            callback_called = True

        listener2.simpleevent.empty_happened.on_event(empty_callback)
        eventer.simpleevent.empty_happened()
        time.sleep(3)
        self.assertTrue(callback_called)

    def test_class_callbacks(self):
        class AdderDoerClass:
            def __init__(self):
                self.val = 0

            def addnumbers(self, a, b):
                self.val = a + b
                return a + b

            def do_it(self):
                self.val += 1


        mgr = AssetManager(HOST, PORT, "test_arena2036", "test_endpoint_1")
        owner = mgr.create_asset(
            "source", res_url + "simple_operation.json", mode=Mode.OWNER
        )
        adder_doer = AdderDoerClass()
        owner.simple_operation.bind_addnumbers(adder_doer.addnumbers)
        owner.simple_operation.bind_do_it(adder_doer.do_it)

        consumer = mgr.create_asset(
            "source", res_url + "simple_operation.json", mode=Mode.CONSUMER
        )

        consumer.simple_operation.addnumbers(a=5, b=3)
        consumer.simple_operation.do_it()

        time.sleep(2)

        self.assertEqual(adder_doer.val, 9)


    def test_integer_numbers(self):
        def add_anything(a, b):
            return a + b


        mgr = AssetManager(HOST, PORT, "test_num_int", "test_num_int")
        asset = mgr.create_asset(
            "test_num_int", res_url + "integer_number_operation.json"
        )

        asset.simple_operation.bind_addintegers(add_anything)
        asset.simple_operation.bind_addnumbers(add_anything)

        asset_proxy = mgr.create_asset_proxy("test_num_int", "test_num_int")
        self.assertEqual(asset_proxy.simple_operation.addintegers(a=2, b=2), 4)
        self.assertEqual(asset_proxy.simple_operation.addnumbers(a=2.0, b=2.0), 4.0)

    def test_different_namespaces(self):

        mgr = AssetManager(HOST, PORT, "test_arena2036", "test_endpoint_1")

        d1_a1 = mgr.create_asset("d1_a1", res_url + "simpleint.json", namespace="d1")
        d1_a2 = mgr.create_asset("d1_a2", res_url + "simpleint.json", namespace="d1")
        d2_a1 = mgr.create_asset("d2_a1", res_url + "simpleint.json", namespace="d2")
        d2_a2 = mgr.create_asset("d2_a2", res_url + "simpleint.json", namespace="d2")
        inherit_a2 = mgr.create_asset("i_a2", res_url + "simpleint.json")


        d1_a1_c = mgr.create_asset(
            "d1_a1", res_url + "simpleint.json", mode=Mode.CONSUMER, namespace="d1"
        )
        d1_a2_c = mgr.create_asset(
            "d1_a2", res_url + "simpleint.json", mode=Mode.CONSUMER, namespace="d1"
        )
        d2_a1_c = mgr.create_asset(
            "d2_a1", res_url + "simpleint.json", mode=Mode.CONSUMER, namespace="d2"
        )
        d2_a2_c = mgr.create_asset(
            "d2_a2", res_url + "simpleint.json", mode=Mode.CONSUMER, namespace="d2"
        )
        explicit_a2_c = mgr.create_asset(
            "i_a2",
            res_url + "simpleint.json",
            mode=Mode.CONSUMER,
            namespace="test_arena2036",
        )


        d1_a1.simpleint.num.value = 42
        d2_a1.simpleint.num.value = 42
        d1_a2.simpleint.num.value = 42
        d2_a2.simpleint.num.value = 42
        inherit_a2.simpleint.num.value = 42

        time.sleep(4)

        self.assertEqual(d1_a1_c.simpleint.num.value, 42)
        self.assertEqual(d1_a2_c.simpleint.num.value, 42)
        self.assertEqual(d2_a1_c.simpleint.num.value, 42)
        self.assertEqual(d2_a2_c.simpleint.num.value, 42)
        self.assertEqual(explicit_a2_c.simpleint.num.value, 42)

    def test_operation_context(self):
        from assets2036py import context


        mgr = AssetManager(HOST, PORT, "test_arena2036", "test_endpoint_1")
        sink = mgr.create_asset(
            "source", res_url + "simple_operation.json", mode=Mode.CONSUMER
        )

        owner = mgr.create_asset(
            "source", res_url + "simple_operation.json", mode=Mode.OWNER
        )


        def do_addnumbers(a, b):
            req_id = context.req_id
            self.assertIsNotNone(req_id)
            return a + b

        owner.simple_operation.bind_addnumbers(do_addnumbers)

        time.sleep(1)

        res = sink.simple_operation.addnumbers(a=5, b=6)
        self.assertEqual(res, 11)

    def test_asset_disconnect_callback(self):
        mgr = AssetManager(HOST, PORT, "test_arena2036", "test_endpoint_1")
        mgr2 = AssetManager(HOST, PORT, "test_arena2036", "test_endpoint_2")
        asset = mgr.create_asset("test_asset", res_url + "simple_prop.json")
        callback_called = False

        def disconnect_callback(*args):
            nonlocal callback_called
            callback_called = True
            logger.debug("DISCONNECTED: %s", args)

        def run_asset():
            for i in range(5):
                asset.simple_prop_json.my_property.value = i
                time.sleep(0.5)
            asset.disconnect()

        Thread(target=run_asset, daemon=True).start()

        asset_proxy = mgr2.create_asset_proxy("test_arena2036", "test_asset")
        asset_proxy.on_disconnect(disconnect_callback)
        for _ in range(20):
            my_prop = asset_proxy.simple_prop_json.my_property.value
            if callback_called:
                break
            time.sleep(0.5)

        self.assertTrue(callback_called)

    def test_is_proxy_asset_online(self):
        mgr = AssetManager(HOST, PORT, "test_arena2036", "test_endpoint_1")
        mgr2 = AssetManager(HOST, PORT, "test_arena2036", "test_endpoint_2")
        asset = mgr.create_asset("test_asset", res_url + "simple_prop.json")
        asset.simple_prop_json.my_property.value = 42

        asset_proxy = mgr2.create_asset_proxy("test_arena2036", "test_asset")
        self.assertTrue(asset_proxy.is_online)
        asset.disconnect()
        time.sleep(1)
        self.assertFalse(asset_proxy.is_online)

    def test_proxy_not_online_exception(self):
        mgr = AssetManager(HOST, PORT, "test_arena2036", "test_endpoint_1")
        asset = mgr.create_asset("test_asset", res_url + "simple_prop.json")
        asset.simple_prop_json.my_property.value = 42
        time.sleep(1)

        def create_proxy():

            mgr2 = AssetManager(HOST, PORT, "test_arena2036", "test_endpoint_2")

            asset_proxy = mgr2.create_asset_proxy("test_arena2036", "test_asset", 3)
            return True

        self.assertTrue(create_proxy())
        asset.disconnect()
        time.sleep(1)
        self.assertRaises(AssetNotOnlineError, create_proxy)

    def test_lazy_load_events(self):
        mgr = AssetManager(HOST, PORT, "test_arena2036", "test_endpoint_1")
        _source = mgr.create_asset("source", res_url + "simpleevent.json")
        sink = mgr.create_asset_proxy("test_arena2036", "source", lazy_loading=True)

        self.assertFalse("bool_happened" in dir(sink.simpleevent))

        # access event name
        _ = sink.simpleevent.bool_happened.name

        self.assertTrue("bool_happened" in dir(sink.simpleevent))


    def test_lazy_load_properties(self):
        mgr = AssetManager(HOST, PORT, "test_arena2036", "test_endpoint_1")
        source = mgr.create_asset("source", res_url + "all_property_model.json")
        sink = mgr.create_asset(
            "source",
            res_url + "all_property_model.json",
            mode=Mode.CONSUMER,
            lazy_loading=True,
        )


        # Set the properties in the source node
        source.all_properties.a_switch.value = True
        source.all_properties.an_int.value = 42
        source.all_properties.a_float.value = 4.2
        source.all_properties.a_text.value = "Hullo"
        source.all_properties.an_object.value = {"x": 5, "y": 6}
        source.all_properties.a_list.value = [1, 2, 3, 4]

        # Check that the sink does not have the properties set up yet
        self.assertFalse("a_switch" in dir(sink.all_properties))
        self.assertFalse("an_int" in dir(sink.all_properties))
        self.assertFalse("a_float" in dir(sink.all_properties))
        self.assertFalse("a_text" in dir(sink.all_properties))
        self.assertFalse("an_object" in dir(sink.all_properties))
        self.assertFalse("a_list" in dir(sink.all_properties))

        # Access all values once
        _ = sink.all_properties.a_switch.value
        _ = sink.all_properties.an_int.value
        _ = sink.all_properties.a_float.value
        _ = sink.all_properties.a_text.value
        _ = sink.all_properties.an_object.value
        _ = sink.all_properties.a_list.value

        # Check that now all properties are available as attributes
        self.assertTrue("a_switch" in dir(sink.all_properties))
        self.assertTrue("an_int" in dir(sink.all_properties))
        self.assertTrue("a_float" in dir(sink.all_properties))
        self.assertTrue("a_text" in dir(sink.all_properties))
        self.assertTrue("an_object" in dir(sink.all_properties))
        self.assertTrue("a_list" in dir(sink.all_properties))

        # Sleep shortly for MQTT to update
        time.sleep(2)

        # Now assert that the properties also have the right value

        self.assertEqual(True, sink.all_properties.a_switch.value)
        self.assertEqual(42, sink.all_properties.an_int.value)
        self.assertAlmostEqual(4.2, sink.all_properties.a_float.value, places=5)
        self.assertEqual("Hullo", sink.all_properties.a_text.value)
        self.assertDictEqual({"x": 5, "y": 6}, sink.all_properties.an_object.value)
        self.assertListEqual([1, 2, 3, 4], sink.all_properties.a_list.value)
        mgr.disconnect()
